/**
 * \file pathTool.cpp
 *
 * Source file for the rednav path tool.
 *
 * \date 2011-01-07
 * \author redouane.boumghar@laas.fr
 * \ingroup rednav
 */

#include <rednav/pathTool.hpp>

namespace jafar {
namespace rednav {

	// Constructor that does nothing but giving calls to tools
	PathTool::PathTool()
	{ // {{{
		// Default values for all parameters
		subsampler = 4.0;
		maxwaypts  = 0;
		pathMaxEntropy = 0.69135;
		pathMaxProba = 0.8;
		cellScale = 1.0;
	} // }}}

	// Tooled Constructor
	PathTool::PathTool(jafar::lgl::NavPath& path, PATH_TRANSFORMATION trsf)
	{ // {{{
		subsampler = 4.0;
		maxwaypts  = 0;
		pathMaxEntropy = 0.69135;
		pathMaxProba = 0.8;
		cellScale = 1.0;
	} // }}}

	//-------------------------- Destructor ---------
	PathTool::~PathTool()
	{ // {{{
	} // }}}

	//-------------------------- Setting Functions  ---------
	void PathTool::setMaxWayPts( int n )
	{//{{{
		maxwaypts = n;
	}//}}}

	void PathTool::setPathMaxEntropy( double pme )
	{//{{{
		pathMaxEntropy = pme;
	}//}}}

	void PathTool::setPathMaxProba( double pmp )
	{//{{{
		pathMaxProba = pmp;
	}//}}}

	void PathTool::setScale( double cellres)
	{//{{{
		cellScale = cellres;
	}//}}}

	void PathTool::setSubSampler( double setsb)
	{//{{{
		subsampler = setsb;
	}//}}}

	//-------------------------- Getting Functions  ---------
	double PathTool::getPathMaxEntropy()
	{//{{{
		return pathMaxEntropy;
	}//}}}

	double PathTool::getPathMaxProba()
	{//{{{
		return pathMaxProba;
	}//}}}

	double PathTool::getSubSampler()
	{//{{{
		return subsampler;
	}//}}}

	//-------------------------- Other Functions  ---------
	bool PathTool::applyTool(const jafar::lgl::NavPath& inpath, jafar::lgl::NavPath& outpath, PATH_TRANSFORMATION ptrans)
	{//{{{
		switch(ptrans)
		{
			case PATH_TWO_POINTS:
				return (twoPoints(inpath, outpath));
				break;
			case PATH_RDV_CROP:
				return (cropPath(inpath, outpath));
			default: // PATH_COMMON_FILTER
				return (subSample(inpath, outpath, subsampler));
		}
		return false;
	}//}}}

	bool PathTool::twoPoints(const jafar::lgl::NavPath& foundPath, jafar::lgl::NavPath& tPath)
	{//{{{
		jafar::lgl::Attributes emptyAtt;
		// Init.
		jafar::lgl::NavPath::const_iterator path_it;
		jafar::lgl::NavPath::const_iterator path_it2;

		// Temporary pointers to the start and goal points
		path_it = foundPath.begin();
		if (path_it == foundPath.end()) { return false; }
		path_it2 = foundPath.end(); --path_it2;
		if (path_it2 == foundPath.end()) { return false; }

		//---------- Prepare two-points ways for Ressac v1
		// Ressac v1 : ressac only able to follow two points for mapping
		//----- Find the farthest points to the line start-goal on each side
		// start-goal vector coordinates
		double vx_sg = (double)(((*path_it2).first).pos_x - ((*path_it).first).pos_x);
		double vy_sg = (double)(((*path_it2).first).pos_y - ((*path_it).first).pos_y);
		// 'Absolute' start coordinates
		double co_sx = (double)((*path_it).first).pos_x;
		double co_sy = (double)((*path_it).first).pos_y;
		// Relative Goal coordinates with transformation Start --> (0,0)
		double co_gx = (double)((*path_it2).first).pos_x - co_sx;
		double co_gy = (double)((*path_it2).first).pos_y - co_sy;
		// Considered path point coordinates
		double co_ax = 0.0;
		double co_ay = 0.0;
		// 'Absolute' On-DirectLine projected points coordinates
		double co_hx = 0.0;
		double co_hy = 0.0;
		// Slope coefficient of the direct line between start and goal points
		bool mslope_vertical = false;
		double mslope   = 0.0;
		double mslope_c = 0.0;
		if (vx_sg != 0.0) 
		{
			mslope = vy_sg / vx_sg;
			mslope_c = co_gx + mslope * co_gy;
		} else {
			mslope_vertical =  true;
		}
		// Distance parameters
		double dist = 0.0;
		double maxdist_a = 0.0;
		double maxdist_b = 0.0;
		// Axis relative signed position
		double sign_a = 0.0; 

		// To tell if we have found at least one point on the positive side
		int plus_ni  = -1;
		double plus_ax = 0.0;
		double plus_ay = 0.0;
		// To tell if we have found at least one point on the negative side
		int minus_ni = -1;
		double minus_ax = 0.0;
		double minus_ay = 0.0;
		
		int N = 0;
		for( path_it  = foundPath.begin(); 
			 path_it != foundPath.end();
			 ++path_it)
		{
			co_ax = (double)((*path_it).first).pos_x ;
			co_ay = (double)((*path_it).first).pos_y ;
			if (mslope_vertical)
			{
				co_hx = co_sx;
				co_hy = co_ay;
			} else {
				// Getting relative position when start --> (0,0)
				co_ax = co_ax - co_sx;
				co_ay = co_ay - co_sy;
				// Calculating relative projection
				co_hx = (co_ax * co_gx + co_ay * co_gy) / mslope_c;
				co_hy = mslope * co_hx;
				// Translating to the main frame
				co_ax = co_ax + co_sx;
				co_ay = co_ay + co_sy;
				co_hx = co_hx + co_sx;
				co_hy = co_hy + co_sy;
			}

			dist = sqrt( (co_hx-co_ax) * (co_hx-co_ax) 
					   + (co_hy-co_ay) * (co_hy-co_ay) ) ;
			// Saving farthest point on its side
			sign_a = (co_hx - co_ax)*(co_gy)
				   + (co_hy - co_ay)*(co_gx);
			if ( sign_a > 0.0 )
			{
				if (dist > maxdist_a)
				{
					maxdist_a = dist;
					plus_ax = co_ax;
					plus_ay = co_ay;
					plus_ni = N;
				}
			} else {
				if (dist >= maxdist_b)
				{
					maxdist_b = dist;
					minus_ax = co_ax;
					minus_ay = co_ay;
					minus_ni = N;
				}
			}

			// Full path index count
			N++;
		}

		//----- Filling a 2-waypoints path
		tPath.clear();
		std::pair<jafar::lgl::NavLocation, jafar::lgl::Attributes> insertor;
		insertor.second = emptyAtt;
		if ( plus_ni >= 0 && minus_ni >= 0)
		{
			// There is a point on each side of the start-goal line
			insertor.first = jafar::lgl::NavLocation(plus_ax, plus_ay);
			tPath.push_back(insertor);
			insertor.first = jafar::lgl::NavLocation(minus_ax, minus_ay);
			tPath.push_back(insertor);
		} else {
			if (plus_ni >=0 && minus_ni == -1)
			{
				insertor.first = jafar::lgl::NavLocation(plus_ax, plus_ay);
				tPath.push_back(insertor);
				path_it = foundPath.begin();
				if (path_it != foundPath.end())
				{
					tPath.push_back((*path_it));
				} else {
					return false;
				}
			} else {
				if (minus_ni >=0 && plus_ni == -1)
				{
					insertor.first = jafar::lgl::NavLocation(minus_ax, minus_ay);
					tPath.push_back(insertor);
					path_it = foundPath.begin();
					if (path_it != foundPath.end())
					{
						tPath.push_back((*path_it));
					} else {
						return false;
					}
				} else {
					// No points found ? quitely impossible :)
					return false;
				}
			}
		}

		return true;
	}//}}}

	bool PathTool::subSample(const jafar::lgl::NavPath& foundPath, jafar::lgl::NavPath& tPath, double subsamp)
	{//{{{
		//------------ variables
		double x_new = 0.0;
		double y_new = 0.0;
		double x_prevpos = 0.0;
		double y_prevpos = 0.0;
		double curvDist = 0.0;
		int N = 0;
		jafar::lgl::NavPath::const_iterator path_it;
		tPath.clear();

		//--- Subsample the input path to give something smoother for the
		//--- obstacle avoider
		for( path_it = foundPath.begin(); 
			 path_it != foundPath.end();
			 ++path_it)
		{ 
			// Correcting coordinates in the superimposed LGL frame
			x_new = (double)((*path_it).first).pos_x;
			x_new = (x_new/2.0)*cellScale;
			y_new = (double)((*path_it).first).pos_y; 
			y_new = (y_new/2.0)*cellScale;
			if (N==0)
			{
				// Not using first point of the trajectory as it is the
				// position of the vehicule itself or the first point in the
				// trajectory. Our reference.
				N++;
				tPath.push_back((*path_it));
			} else {
				// Resampling trajectory to a more loose one
				x_prevpos = x_prevpos - x_new;
				x_prevpos = x_prevpos*x_prevpos;
				y_prevpos = y_prevpos - y_new;
				y_prevpos = y_prevpos*y_prevpos;
				curvDist = curvDist + sqrt(x_prevpos + y_prevpos);
				if ( curvDist >= subsamp)
				{
					// Adding points to the transformed path
					tPath.push_back((*path_it));
					curvDist = 0.0;
					N++;
				} // else not adding it. 
			}	

			if ( (maxwaypts > 0) && (N > maxwaypts) )
			{
				// No more waypoints needed.
				return true;
			}
			x_prevpos = x_new;
			y_prevpos = y_new;
		}

		if (N == 0)
		{
			return false;
		}
		return true;
	}//}}}

	bool PathTool::cropPath(const jafar::lgl::NavPath& srcPath, jafar::lgl::NavPath& outPath)
	{//{{{
		if ( pathMaxEntropy <= 0.0 && pathMaxProba <= 0.0)
		{
			// No crop possible.
			return false;
		}
		jafar::lgl::NavPath::const_iterator path_it;
		int n = 0;
		outPath.clear();
		std::pair<jafar::lgl::NavLocation, jafar::lgl::Attributes> insertor;
		for(path_it = srcPath.begin(); (path_it != srcPath.end()); ++path_it)
		{ 
			// Testing if limits (pathMaxEntropy and pathMaxProba) are reach OR if maxwaypts is
			// set to more than zero crop the path to that maximum allowed
			// number of waypoints.
			if (   (pathMaxEntropy > 0.0 && (((*path_it).second).infotropy[jafar::lgl::T_SYMBOL])   >= pathMaxEntropy)
				|| (pathMaxProba   > 0.0 && (((*path_it).second).certainty[jafar::lgl::T_OBSTACLE]) >= pathMaxProba)
				|| (maxwaypts > 0 && n > maxwaypts) )
			{
				// Working on copy
				//srcPath.erase(path_it, srcPath.end());
				return true;
			}
			insertor.first  = path_it->first;
			insertor.second = path_it->second;
			outPath.push_back(insertor);
			//outPath.push_back(std::make_pair(path_it->first, path_it->second));
			n++;
		}
		// No cropping has been done
		return false;
	}//}}}

	bool PathTool::cropPath(double vent, double vproba, jafar::lgl::NavPath& navpath)
	{//{{{
		if ( vent <= 0.0 && vproba <= 0.0)
		{
			// No crop possible.
			return false;
		}
		jafar::lgl::NavPath::iterator path_it;
		int n = 0;
		for(path_it = navpath.begin(); (path_it != navpath.end()); ++path_it)
		{ 
			// Testing if limits (vent and vproba) are reach OR if maxwaypts is
			// set to more than zero crop the path to that maximum allowed
			// number of waypoints.
			if (   (vent   > 0.0 && (((*path_it).second).infotropy[jafar::lgl::T_SYMBOL])   >= vent )
				|| (vproba > 0.0 && (((*path_it).second).certainty[jafar::lgl::T_OBSTACLE]) >= vproba)
				|| (maxwaypts > 0 && n > maxwaypts) )
			{
				navpath.erase(path_it, navpath.end());
				return true;
			}
			n++;
		}
		// No cropping has been done
		return false;
	}//}}}

}} // end of namespaces
